package org.apache.bcel.generic;

import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import org.apache.bcel.Const;
import org.apache.bcel.classfile.Constant;
import org.apache.bcel.util.ByteSequence;
import org.apache.commons.lang3.ArrayUtils;

public class InstructionList implements Iterable<InstructionHandle> {
   private InstructionHandle start;
   private InstructionHandle end;
   private int length;
   private int[] bytePositions;
   private List<InstructionListObserver> observers;

   public static InstructionHandle findHandle(InstructionHandle[] ihs, int[] pos, int count, int target) {
      int l = 0;
      int r = count - 1;

      do {
         int i = l + r >>> 1;
         int j = pos[i];
         if (j == target) {
            return ihs[i];
         }

         if (target < j) {
            r = i - 1;
         } else {
            l = i + 1;
         }
      } while(l <= r);

      return null;
   }

   public InstructionList() {
   }

   public InstructionList(BranchInstruction i) {
      this.append(i);
   }

   public InstructionList(byte[] code) {
      int count = 0;

      int[] pos;
      InstructionHandle[] ihs;
      int target;
      try {
         ByteSequence bytes = new ByteSequence(code);
         Throwable var6 = null;

         try {
            ihs = new InstructionHandle[code.length];

            for(pos = new int[code.length]; bytes.available() > 0; ++count) {
               target = bytes.getIndex();
               pos[count] = target;
               Instruction i = Instruction.readInstruction(bytes);
               Object ih;
               if (i instanceof BranchInstruction) {
                  ih = this.append((BranchInstruction)i);
               } else {
                  ih = this.append(i);
               }

               ((InstructionHandle)ih).setPosition(target);
               ihs[count] = (InstructionHandle)ih;
            }
         } catch (Throwable var18) {
            var6 = var18;
            throw var18;
         } finally {
            if (bytes != null) {
               if (var6 != null) {
                  try {
                     bytes.close();
                  } catch (Throwable var17) {
                     var6.addSuppressed(var17);
                  }
               } else {
                  bytes.close();
               }
            }

         }
      } catch (IOException var20) {
         IOException e = var20;
         throw new ClassGenException(e.toString(), e);
      }

      this.bytePositions = Arrays.copyOf(pos, count);

      for(int i = 0; i < count; ++i) {
         if (ihs[i] instanceof BranchHandle) {
            BranchInstruction bi = (BranchInstruction)ihs[i].getInstruction();
            target = bi.getPosition() + bi.getIndex();
            InstructionHandle ih = findHandle(ihs, pos, count, target);
            if (ih == null) {
               throw new ClassGenException("Couldn't find target for branch: " + bi);
            }

            bi.setTarget(ih);
            if (bi instanceof Select) {
               Select s = (Select)bi;
               int[] indices = s.getIndices();

               for(int j = 0; j < indices.length; ++j) {
                  target = bi.getPosition() + indices[j];
                  ih = findHandle(ihs, pos, count, target);
                  if (ih == null) {
                     throw new ClassGenException("Couldn't find target for switch: " + bi);
                  }

                  s.setTarget(j, ih);
               }
            }
         }
      }

   }

   public InstructionList(CompoundInstruction c) {
      this.append(c.getInstructionList());
   }

   public InstructionList(Instruction i) {
      this.append(i);
   }

   public void addObserver(InstructionListObserver o) {
      if (this.observers == null) {
         this.observers = new ArrayList();
      }

      this.observers.add(o);
   }

   public BranchHandle append(BranchInstruction i) {
      BranchHandle ih = BranchHandle.getBranchHandle(i);
      this.append((InstructionHandle)ih);
      return ih;
   }

   public InstructionHandle append(CompoundInstruction c) {
      return this.append(c.getInstructionList());
   }

   public InstructionHandle append(Instruction i) {
      InstructionHandle ih = InstructionHandle.getInstructionHandle(i);
      this.append(ih);
      return ih;
   }

   public InstructionHandle append(Instruction i, CompoundInstruction c) {
      return this.append(i, c.getInstructionList());
   }

   public InstructionHandle append(Instruction i, Instruction j) {
      return this.append(i, new InstructionList(j));
   }

   public InstructionHandle append(Instruction i, InstructionList il) {
      InstructionHandle ih;
      if ((ih = this.findInstruction2(i)) == null) {
         throw new ClassGenException("Instruction " + i + " is not contained in this list.");
      } else {
         return this.append(ih, il);
      }
   }

   private void append(InstructionHandle ih) {
      if (this.isEmpty()) {
         this.start = this.end = ih;
         ih.setNext(ih.setPrev((InstructionHandle)null));
      } else {
         this.end.setNext(ih);
         ih.setPrev(this.end);
         ih.setNext((InstructionHandle)null);
         this.end = ih;
      }

      ++this.length;
   }

   public BranchHandle append(InstructionHandle ih, BranchInstruction i) {
      BranchHandle bh = BranchHandle.getBranchHandle(i);
      InstructionList il = new InstructionList();
      il.append((InstructionHandle)bh);
      this.append(ih, il);
      return bh;
   }

   public InstructionHandle append(InstructionHandle ih, CompoundInstruction c) {
      return this.append(ih, c.getInstructionList());
   }

   public InstructionHandle append(InstructionHandle ih, Instruction i) {
      return this.append(ih, new InstructionList(i));
   }

   public InstructionHandle append(InstructionHandle ih, InstructionList il) {
      if (il == null) {
         throw new ClassGenException("Appending null InstructionList");
      } else if (il.isEmpty()) {
         return ih;
      } else {
         InstructionHandle next = ih.getNext();
         InstructionHandle ret = il.start;
         ih.setNext(il.start);
         il.start.setPrev(ih);
         il.end.setNext(next);
         if (next != null) {
            next.setPrev(il.end);
         } else {
            this.end = il.end;
         }

         this.length += il.length;
         il.clear();
         return ret;
      }
   }

   public InstructionHandle append(InstructionList il) {
      if (il == null) {
         throw new ClassGenException("Appending null InstructionList");
      } else if (il.isEmpty()) {
         return null;
      } else if (this.isEmpty()) {
         this.start = il.start;
         this.end = il.end;
         this.length = il.length;
         il.clear();
         return this.start;
      } else {
         return this.append(this.end, il);
      }
   }

   private void clear() {
      this.start = this.end = null;
      this.length = 0;
   }

   public boolean contains(Instruction i) {
      return this.findInstruction1(i) != null;
   }

   public boolean contains(InstructionHandle i) {
      if (i == null) {
         return false;
      } else {
         for(InstructionHandle ih = this.start; ih != null; ih = ih.getNext()) {
            if (ih == i) {
               return true;
            }
         }

         return false;
      }
   }

   public InstructionList copy() {
      Map<InstructionHandle, InstructionHandle> map = new HashMap();
      InstructionList il = new InstructionList();

      InstructionHandle ih;
      Instruction i;
      for(ih = this.start; ih != null; ih = ih.getNext()) {
         Instruction i = ih.getInstruction();
         i = i.copy();
         if (i instanceof BranchInstruction) {
            map.put(ih, il.append((BranchInstruction)i));
         } else {
            map.put(ih, il.append(i));
         }
      }

      ih = this.start;

      for(InstructionHandle ch = il.start; ih != null; ch = ch.getNext()) {
         i = ih.getInstruction();
         Instruction c = ch.getInstruction();
         if (i instanceof BranchInstruction) {
            BranchInstruction bi = (BranchInstruction)i;
            BranchInstruction bc = (BranchInstruction)c;
            InstructionHandle itarget = bi.getTarget();
            bc.setTarget((InstructionHandle)map.get(itarget));
            if (bi instanceof Select) {
               InstructionHandle[] itargets = ((Select)bi).getTargets();
               InstructionHandle[] ctargets = ((Select)bc).getTargets();

               for(int j = 0; j < itargets.length; ++j) {
                  ctargets[j] = (InstructionHandle)map.get(itargets[j]);
               }
            }
         }

         ih = ih.getNext();
      }

      return il;
   }

   public void delete(Instruction i) throws TargetLostException {
      InstructionHandle ih;
      if ((ih = this.findInstruction1(i)) == null) {
         throw new ClassGenException("Instruction " + i + " is not contained in this list.");
      } else {
         this.delete(ih);
      }
   }

   public void delete(Instruction from, Instruction to) throws TargetLostException {
      InstructionHandle fromIh;
      if ((fromIh = this.findInstruction1(from)) == null) {
         throw new ClassGenException("Instruction " + from + " is not contained in this list.");
      } else {
         InstructionHandle toIh;
         if ((toIh = this.findInstruction2(to)) == null) {
            throw new ClassGenException("Instruction " + to + " is not contained in this list.");
         } else {
            this.delete(fromIh, toIh);
         }
      }
   }

   public void delete(InstructionHandle ih) throws TargetLostException {
      this.remove(ih.getPrev(), ih.getNext());
   }

   public void delete(InstructionHandle from, InstructionHandle to) throws TargetLostException {
      this.remove(from.getPrev(), to.getNext());
   }

   public void dispose() {
      for(InstructionHandle ih = this.end; ih != null; ih = ih.getPrev()) {
         ih.dispose();
      }

      this.clear();
   }

   public InstructionHandle findHandle(int pos) {
      int[] positions = this.bytePositions;
      InstructionHandle ih = this.start;

      for(int i = 0; i < this.length; ++i) {
         if (positions[i] == pos) {
            return ih;
         }

         ih = ih.getNext();
      }

      return null;
   }

   private InstructionHandle findInstruction1(Instruction i) {
      for(InstructionHandle ih = this.start; ih != null; ih = ih.getNext()) {
         if (ih.getInstruction() == i) {
            return ih;
         }
      }

      return null;
   }

   private InstructionHandle findInstruction2(Instruction i) {
      for(InstructionHandle ih = this.end; ih != null; ih = ih.getPrev()) {
         if (ih.getInstruction() == i) {
            return ih;
         }
      }

      return null;
   }

   public byte[] getByteCode() {
      this.setPositions();
      ByteArrayOutputStream b = new ByteArrayOutputStream();
      DataOutputStream out = new DataOutputStream(b);

      try {
         for(InstructionHandle ih = this.start; ih != null; ih = ih.getNext()) {
            Instruction i = ih.getInstruction();
            i.dump(out);
         }

         out.flush();
         return b.toByteArray();
      } catch (IOException var5) {
         IOException e = var5;
         System.err.println(e);
         return ArrayUtils.EMPTY_BYTE_ARRAY;
      }
   }

   public InstructionHandle getEnd() {
      return this.end;
   }

   public InstructionHandle[] getInstructionHandles() {
      InstructionHandle[] ihs = new InstructionHandle[this.length];
      InstructionHandle ih = this.start;

      for(int i = 0; i < this.length; ++i) {
         ihs[i] = ih;
         ih = ih.getNext();
      }

      return ihs;
   }

   public int[] getInstructionPositions() {
      return this.bytePositions;
   }

   public Instruction[] getInstructions() {
      List<Instruction> instructions = new ArrayList();

      try {
         ByteSequence bytes = new ByteSequence(this.getByteCode());
         Throwable var3 = null;

         try {
            while(bytes.available() > 0) {
               instructions.add(Instruction.readInstruction(bytes));
            }
         } catch (Throwable var13) {
            var3 = var13;
            throw var13;
         } finally {
            if (bytes != null) {
               if (var3 != null) {
                  try {
                     bytes.close();
                  } catch (Throwable var12) {
                     var3.addSuppressed(var12);
                  }
               } else {
                  bytes.close();
               }
            }

         }
      } catch (IOException var15) {
         IOException e = var15;
         throw new ClassGenException(e.toString(), e);
      }

      return (Instruction[])instructions.toArray(Instruction.EMPTY_ARRAY);
   }

   public int getLength() {
      return this.length;
   }

   public InstructionHandle getStart() {
      return this.start;
   }

   public BranchHandle insert(BranchInstruction i) {
      BranchHandle ih = BranchHandle.getBranchHandle(i);
      this.insert((InstructionHandle)ih);
      return ih;
   }

   public InstructionHandle insert(CompoundInstruction c) {
      return this.insert(c.getInstructionList());
   }

   public InstructionHandle insert(Instruction i) {
      InstructionHandle ih = InstructionHandle.getInstructionHandle(i);
      this.insert(ih);
      return ih;
   }

   public InstructionHandle insert(Instruction i, CompoundInstruction c) {
      return this.insert(i, c.getInstructionList());
   }

   public InstructionHandle insert(Instruction i, Instruction j) {
      return this.insert(i, new InstructionList(j));
   }

   public InstructionHandle insert(Instruction i, InstructionList il) {
      InstructionHandle ih;
      if ((ih = this.findInstruction1(i)) == null) {
         throw new ClassGenException("Instruction " + i + " is not contained in this list.");
      } else {
         return this.insert(ih, il);
      }
   }

   private void insert(InstructionHandle ih) {
      if (this.isEmpty()) {
         this.start = this.end = ih;
         ih.setNext(ih.setPrev((InstructionHandle)null));
      } else {
         this.start.setPrev(ih);
         ih.setNext(this.start);
         ih.setPrev((InstructionHandle)null);
         this.start = ih;
      }

      ++this.length;
   }

   public BranchHandle insert(InstructionHandle ih, BranchInstruction i) {
      BranchHandle bh = BranchHandle.getBranchHandle(i);
      InstructionList il = new InstructionList();
      il.append((InstructionHandle)bh);
      this.insert(ih, il);
      return bh;
   }

   public InstructionHandle insert(InstructionHandle ih, CompoundInstruction c) {
      return this.insert(ih, c.getInstructionList());
   }

   public InstructionHandle insert(InstructionHandle ih, Instruction i) {
      return this.insert(ih, new InstructionList(i));
   }

   public InstructionHandle insert(InstructionHandle ih, InstructionList il) {
      if (il == null) {
         throw new ClassGenException("Inserting null InstructionList");
      } else if (il.isEmpty()) {
         return ih;
      } else {
         InstructionHandle prev = ih.getPrev();
         InstructionHandle ret = il.start;
         ih.setPrev(il.end);
         il.end.setNext(ih);
         il.start.setPrev(prev);
         if (prev != null) {
            prev.setNext(il.start);
         } else {
            this.start = il.start;
         }

         this.length += il.length;
         il.clear();
         return ret;
      }
   }

   public InstructionHandle insert(InstructionList il) {
      if (this.isEmpty()) {
         this.append(il);
         return this.start;
      } else {
         return this.insert(this.start, il);
      }
   }

   public boolean isEmpty() {
      return this.start == null;
   }

   public Iterator<InstructionHandle> iterator() {
      return new Iterator<InstructionHandle>() {
         private InstructionHandle ih;

         {
            this.ih = InstructionList.this.start;
         }

         public boolean hasNext() {
            return this.ih != null;
         }

         public InstructionHandle next() throws NoSuchElementException {
            if (this.ih == null) {
               throw new NoSuchElementException();
            } else {
               InstructionHandle i = this.ih;
               this.ih = this.ih.getNext();
               return i;
            }
         }

         public void remove() {
            throw new UnsupportedOperationException();
         }
      };
   }

   public void move(InstructionHandle ih, InstructionHandle target) {
      this.move(ih, ih, target);
   }

   public void move(InstructionHandle start, InstructionHandle end, InstructionHandle target) {
      if (start != null && end != null) {
         if (target != start && target != end) {
            InstructionHandle prev;
            for(prev = start; prev != end.getNext(); prev = prev.getNext()) {
               if (prev == null) {
                  throw new ClassGenException("Invalid range: From " + start + " to " + end);
               }

               if (prev == target) {
                  throw new ClassGenException("Invalid range: From " + start + " to " + end + " contains target " + target);
               }
            }

            prev = start.getPrev();
            InstructionHandle next = end.getNext();
            if (prev != null) {
               prev.setNext(next);
            } else {
               this.start = next;
            }

            if (next != null) {
               next.setPrev(prev);
            } else {
               this.end = prev;
            }

            start.setPrev(end.setNext((InstructionHandle)null));
            if (target == null) {
               if (this.start != null) {
                  this.start.setPrev(end);
               }

               end.setNext(this.start);
               this.start = start;
            } else {
               next = target.getNext();
               target.setNext(start);
               start.setPrev(target);
               end.setNext(next);
               if (next != null) {
                  next.setPrev(end);
               } else {
                  this.end = end;
               }
            }

         } else {
            throw new ClassGenException("Invalid range: From " + start + " to " + end + " contains target " + target);
         }
      } else {
         throw new ClassGenException("Invalid null handle: From " + start + " to " + end);
      }
   }

   public void redirectBranches(InstructionHandle oldTarget, InstructionHandle newTarget) {
      for(InstructionHandle ih = this.start; ih != null; ih = ih.getNext()) {
         Instruction i = ih.getInstruction();
         if (i instanceof BranchInstruction) {
            BranchInstruction b = (BranchInstruction)i;
            InstructionHandle target = b.getTarget();
            if (target == oldTarget) {
               b.setTarget(newTarget);
            }

            if (b instanceof Select) {
               InstructionHandle[] targets = ((Select)b).getTargets();

               for(int j = 0; j < targets.length; ++j) {
                  if (targets[j] == oldTarget) {
                     ((Select)b).setTarget(j, newTarget);
                  }
               }
            }
         }
      }

   }

   public void redirectExceptionHandlers(CodeExceptionGen[] exceptions, InstructionHandle oldTarget, InstructionHandle newTarget) {
      CodeExceptionGen[] var4 = exceptions;
      int var5 = exceptions.length;

      for(int var6 = 0; var6 < var5; ++var6) {
         CodeExceptionGen exception = var4[var6];
         if (exception.getStartPC() == oldTarget) {
            exception.setStartPC(newTarget);
         }

         if (exception.getEndPC() == oldTarget) {
            exception.setEndPC(newTarget);
         }

         if (exception.getHandlerPC() == oldTarget) {
            exception.setHandlerPC(newTarget);
         }
      }

   }

   public void redirectLocalVariables(LocalVariableGen[] lg, InstructionHandle oldTarget, InstructionHandle newTarget) {
      LocalVariableGen[] var4 = lg;
      int var5 = lg.length;

      for(int var6 = 0; var6 < var5; ++var6) {
         LocalVariableGen element = var4[var6];
         InstructionHandle start = element.getStart();
         InstructionHandle end = element.getEnd();
         if (start == oldTarget) {
            element.setStart(newTarget);
         }

         if (end == oldTarget) {
            element.setEnd(newTarget);
         }
      }

   }

   private void remove(InstructionHandle prev, InstructionHandle next) throws TargetLostException {
      InstructionHandle first;
      InstructionHandle last;
      if (prev == null && next == null) {
         first = this.start;
         last = this.end;
         this.start = this.end = null;
      } else {
         if (prev == null) {
            first = this.start;
            this.start = next;
         } else {
            first = prev.getNext();
            prev.setNext(next);
         }

         if (next == null) {
            last = this.end;
            this.end = prev;
         } else {
            last = next.getPrev();
            next.setPrev(prev);
         }
      }

      first.setPrev((InstructionHandle)null);
      last.setNext((InstructionHandle)null);
      List<InstructionHandle> targetVec = new ArrayList();

      for(InstructionHandle ih = first; ih != null; ih = ih.getNext()) {
         ih.getInstruction().dispose();
      }

      StringBuilder buf = new StringBuilder("{ ");

      for(InstructionHandle ih = first; ih != null; ih = next) {
         next = ih.getNext();
         --this.length;
         if (ih.hasTargeters()) {
            targetVec.add(ih);
            buf.append(ih.toString(true)).append(" ");
            ih.setNext(ih.setPrev((InstructionHandle)null));
         } else {
            ih.dispose();
         }
      }

      buf.append("}");
      if (!targetVec.isEmpty()) {
         throw new TargetLostException((InstructionHandle[])targetVec.toArray(InstructionHandle.EMPTY_ARRAY), buf.toString());
      }
   }

   public void removeObserver(InstructionListObserver o) {
      if (this.observers != null) {
         this.observers.remove(o);
      }

   }

   public void replaceConstantPool(ConstantPoolGen oldCp, ConstantPoolGen newCp) {
      for(InstructionHandle ih = this.start; ih != null; ih = ih.getNext()) {
         Instruction i = ih.getInstruction();
         if (i instanceof CPInstruction) {
            CPInstruction ci = (CPInstruction)i;
            Constant c = oldCp.getConstant(ci.getIndex());
            ci.setIndex(newCp.addConstant(c, oldCp));
         }
      }

   }

   public void setPositions() {
      this.setPositions(false);
   }

   public void setPositions(boolean check) {
      int maxAdditionalBytes = 0;
      int additionalBytes = 0;
      int index = 0;
      int count = 0;
      int[] pos = new int[this.length];
      InstructionHandle ih;
      Instruction i;
      if (check) {
         for(ih = this.start; ih != null; ih = ih.getNext()) {
            i = ih.getInstruction();
            if (i instanceof BranchInstruction) {
               Instruction inst = ((BranchInstruction)i).getTarget().getInstruction();
               if (!this.contains(inst)) {
                  throw new ClassGenException("Branch target of " + Const.getOpcodeName(i.getOpcode()) + ":" + inst + " not in instruction list");
               }

               if (i instanceof Select) {
                  InstructionHandle[] targets = ((Select)i).getTargets();
                  InstructionHandle[] var11 = targets;
                  int var12 = targets.length;

                  for(int var13 = 0; var13 < var12; ++var13) {
                     InstructionHandle target = var11[var13];
                     inst = target.getInstruction();
                     if (!this.contains(inst)) {
                        throw new ClassGenException("Branch target of " + Const.getOpcodeName(i.getOpcode()) + ":" + inst + " not in instruction list");
                     }
                  }
               }

               if (!(ih instanceof BranchHandle)) {
                  throw new ClassGenException("Branch instruction " + Const.getOpcodeName(i.getOpcode()) + ":" + inst + " not contained in BranchHandle.");
               }
            }
         }
      }

      for(ih = this.start; ih != null; ih = ih.getNext()) {
         i = ih.getInstruction();
         ih.setPosition(index);
         pos[count++] = index;
         switch (i.getOpcode()) {
            case 167:
            case 168:
               maxAdditionalBytes += 2;
            case 169:
            default:
               break;
            case 170:
            case 171:
               maxAdditionalBytes += 3;
         }

         index += i.getLength();
      }

      for(ih = this.start; ih != null; ih = ih.getNext()) {
         additionalBytes += ih.updatePosition(additionalBytes, maxAdditionalBytes);
      }

      count = 0;
      index = 0;

      for(ih = this.start; ih != null; ih = ih.getNext()) {
         i = ih.getInstruction();
         ih.setPosition(index);
         pos[count++] = index;
         index += i.getLength();
      }

      this.bytePositions = new int[count];
      System.arraycopy(pos, 0, this.bytePositions, 0, count);
   }

   public int size() {
      return this.length;
   }

   public String toString() {
      return this.toString(true);
   }

   public String toString(boolean verbose) {
      StringBuilder buf = new StringBuilder();

      for(InstructionHandle ih = this.start; ih != null; ih = ih.getNext()) {
         buf.append(ih.toString(verbose)).append("\n");
      }

      return buf.toString();
   }

   public void update() {
      if (this.observers != null) {
         Iterator var1 = this.observers.iterator();

         while(var1.hasNext()) {
            InstructionListObserver observer = (InstructionListObserver)var1.next();
            observer.notify(this);
         }
      }

   }
}
